// g++ -g -Wall -std=c++11 shell_skel.cpp -o monshell.exx

#if (!defined(__cplusplus) || (__cplusplus != 201103L))
#error Il faut utiliser c++11. Apprenez à configurer votre IDE favori ou a lire les informations en début de fichier.
#endif

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <cstring>
#include <cerrno>
#include <regex>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>

using namespace std;

/************************************
 *
 *  Traîtement d'erreur
 *
 ************************************/
/**
 * Fonction générique de traîtement d'erreur.
 * Elle n'est pas très intéressante à regarder,
 * mais utile pour simplifier le code dans la suite.
 */
void
__error_(const char *file, const int ligne, const std::string &msg, bool cond, int errnum, bool exception) {
    
  if (cond) {
    // il y a bien une erreur
    std::ostringstream c;
    if (exception) {
      c <<  "\033[31;11mERREUR\033[00m";
    } else {
      c <<  "\033[34;11mWARNING\033[00m";
    }
    c << " fichier " << file << " ligne " <<  ligne << " : ";
    if (errnum != 0) {
      std::system_error err = std::system_error(errno, std::system_category(), msg);
      c << err.what() << std::endl;
      std::cerr << c.str();
      if (exception) {
	throw err;
      }
    } else {
      std::runtime_error err = std::runtime_error(msg);
      c << err.what() << std::endl;
      std::cerr << c.str();
      if (exception) {
	throw err;
      }
    }
  }
}
/**
 *  @brief teste errnum et envoie une exception si errnum
 *  est un code d'erreur différent de 0
 */
#define exit_error(msg, cond, errnum) __error_(__FILE__, __LINE__, (msg), cond, errnum, true)
/**
 * @brief teste errnum et affiche un message si errnum
 * est un code d'erreur différent de 0
 */
#define warning_error(msg, cond, errnum) __error_(__FILE__, __LINE__, (msg), cond, errnum, false)


/**************************************
 *
 * Fonction utilitaire (pas forcément utile à regarder)
 *
 */
/**
   @brief fonction utilitaire pour découper une chaîne de caractères en morceaux
   @param s : la chaîne à découper
   @param e : la regex pour détecter les séparateurs
   @return un tableau de chaînes de caractères terminé par null.
   @see delTabChar;

   Le tableau retourné a été réservé par malloc et doit être libéré par fonction
   delTabChar. Attention la fonction ignore les tockens vides.
   
 */
char **splitToChar(string s,const std::regex &e) {
  unsigned int i;
  std::vector<string> spl;
  std::smatch m;
  // découpage
  while (std::regex_search(s, m, e)) {
    string pre = m.prefix().str();
    if (pre.size() != 0) {
      spl.push_back(pre);
    }
    s = m.suffix();
  }
  if (s.size() != 0) {
    spl.push_back(s);
  }

  // conversion 
  char **res = new char*[spl.size()+1];
  for (i=0; i<spl.size(); i++) {
    
    res[i]=new char[spl[i].size()+1];
    strcpy(res[i], spl[i].c_str());
  }
  res[i] = NULL;

  return res;
}

void delTabChar(char **t) {
  int i;
  for (i=0; t[i]!= NULL; i++) {
    delete[] t[i];
    t[i] = NULL;
  }
  delete[] t;
}

/****************************************
 *
 * Partie importante
 *
 */

/**
   @brief fonction pour simuler l'exécution d'un programme.
   Elle prend les mêmes paramêtres que execvp(), la fonction qu'il faut appeler.
   @param file : le nom du programme à éxécuter
   @param argv : un tableau de chaînes de caractères contenant les arguments
   	         et terminé par NULL
*/
int executevp(const char *file, char *const argv[]) {
  char *const *pt;
  
  cout << "Le shell doit éxécuter la commande '" << file << "' avec les arguments :" << endl;
    
  pt = argv;
  while (*pt != NULL) {
    cout << "\t- '" << *pt << "'" << endl;
    pt++;
  }
  return 0;
}

/**
   @brief fonction pour attendre un fils dont on connaît le pid
   @param pid : le pid du fils à attendre
   @return le numero du fils qui s'est arrété
*/
int attente_fils(int pid) {
  int status;
  pid_t res;

  res = waitpid(pid, &status, 0);
  exit_error("waitpid", res==-1, errno);

  if (WIFEXITED(status)) {
    fprintf(stdout, "la commande s'est terminée en renvoyant %d\n", WEXITSTATUS(status));
    return res;
  }
  
  if (WIFSIGNALED(status)) {
    fprintf(stdout, "la commande s'est terminée suite au signal %d\n", WTERMSIG(status));
    return res;
  }
  
  // on sort car normalement on ne doit pas passer là
  exit_error("ceci ne devrait pas arriver\n", 1, 0);
  return 0;
}

pid_t pid_attendu;
void traitement_signal_fils (int sig) {
  int res;

  //  fprintf(stdout, "coucou\n");
  res = attente_fils(-1);
  cout << "pid du fils" << res << endl;
  // res est le numéro du fils qui s'est terminé
  if (pid_attendu == res) {
    pid_attendu = 0;
  }
}

typedef void (*sighandler_t)(int);

int main(int argc, char *argv[]) {
  sighandler_t old_h;
  
  old_h = signal(SIGCHLD, traitement_signal_fils);
  exit_error("sigaction", old_h == SIG_ERR, errno);
  
  while (1) {
    int attend;

    string text_com;

    //    text_com = readline("Entrez une commande>");
    cout << "Entrez une commande>" << std::flush;
    
    // lecture de la ligne
    getline(cin, text_com);
    if (cin.eof()) {
      cout << "goodbye" << endl;
      break;
    }
    if (cin.fail()) {
      cout << "Le shell à un problème" << endl;
      break;
    }


    // on normalise la ligne en supprimant les espaces doubles
    // en début ou fin de ligne
    std::regex sps("[ \t\n]+");
    text_com = std::regex_replace(text_com, sps, " ");
    text_com = std::regex_replace(text_com, std::regex("^ +"), "");
    text_com = std::regex_replace(text_com, std::regex(" +$"), "");

    cerr << "Lecture de la commande \"" << text_com << "\"" << endl;

    if (text_com == "fin") {
      cout << "Bye." << endl;
      break;
    }

    if (text_com == "") {
      cerr << "La comande est vide." << endl;
      continue;
    }
      
    if (text_com.back() == '&') {
      // Il y a un & à la fin
      // On doit faire une commande en tâche de fond.
      attend = false;
      text_com.pop_back();
      text_com = std::regex_replace(text_com, std::regex(" +$"), "");
    } else {
      // Le shell devra attendre la fin de la commande
      attend = true;
    }

    // Y a t'il un pipe | dans la commande
    if (!std::regex_match(text_com, std::regex(".*[|].*"))) {
      // Non il n'y en a pas
      // Conversion de la commande en tableau utilisable
      // par les fonctions système
      char **arg_com = splitToChar(text_com, std::regex(" "));
      
      // ######################################
      // Code à modifier ici
      // ######################################
      /**
	 @todo partie 1
	 vous devez appeler la fonction execvp() qui prend
	 les mêmes arguments que executevp().
	 Attention, vous devez faire en sorte que
	 - shell exécute la commande en premier plan si attend == true
	 - en tâche de fond sinon
	 - continue après l'appel à cette fonction
      */
      //      executevp(arg_com[0], arg_com);
      pid_t pid_com = fork();
      exit_error("fork", pid_com==-1, errno);
      if (pid_com == 0) {
	// le fils
	cerr << "Lancement de " << text_com << endl;
	execvp(arg_com[0], arg_com);
	cerr << "Erreur avec la commande " << text_com << " car : " << strerror(errno) << endl;
	exit(2);
      }
      
      if (attend) {
	pid_attendu = pid_com;
	while (pid_com == pid_attendu) {
	  pause();
	}
      } else {
	// rien a faire
	// cerr << "En tâche de fond je ne sait pas faire\n";
	// attente_fils(pid_com);
      }
      
      // libération du tableau arg_com
      delTabChar(arg_com);
    } else {
      // il y a au moins un pipe. On teste s'il n'y en a qu'un

      std::smatch mpipe;
      if (!std::regex_search(text_com, mpipe,
			    std::regex("^([^|]+)\b*[|]\b*([^|]+)$"))) {
	// Il y a plus de 1 pipe, on ne traite pas
	cerr << "La commande " << text_com << " contient plusieurs pipes, on ne la traite pas" << endl;
	continue;
      }
      // Il n'y a qu'un seul pipe, donc 2 commandes à faire
      char **arg_com1 = splitToChar(mpipe[1].str(), std::regex("[ \t]"));
      char **arg_com2 = splitToChar(mpipe[2].str(), std::regex("[ \t]"));
      
      // ######################################
      // Code à modifier ici
      // ######################################
      // vous devez appeler la fonction execvp() avec les
      // arguments identiques que pour executevp()
      // de manière à ce que le shell exécute les deux commandes
      // avec un pipe entre elles
      executevp(arg_com1[0], arg_com1);
      fprintf(stdout, "Dont la sortie est redirigée vers\n");
      executevp(arg_com2[0], arg_com2);
      
      if (attend) {
	fprintf(stdout, "Au premier plan\n");
	/// @todo attendre
      } else {
	fprintf(stdout, "En tâche de fond\n");	  
	/// @todo ne pas attendre
      }
      
    }
    
  }   

  return 0;
}
